<!DOCTYPE html>
<html>
    <head>
        <!-- This presentation is generated by Vitality {{ version }}. -->
        <title>
            {{ data.title | default("Presentation", true) }}
        </title>
        <style>
            {% include "styles.css" %}
        </style>
        {% for font in data.fonts %}
            <link href="https://fonts.googleapis.com/css?family={{ font }}&display=swap" rel="stylesheet">
        {% endfor %}
        {% if remote_d3 %}
            <script src="https://d3js.org/d3.v5.min.js"></script>
        {% else %}
            <script>{% include "d3.js" %}</script>
        {% endif %}
        <script>
            {% include "data.js" %}
            {% include "vitality.js" %}
        </script>
    </head>
    <body>
    </body>
</html>
