/* Centering class for users to center HTML content */
.center {
    position: absolute;
    top: 50%;
    left: 50%;
    transform: translate(-50%, -50%);
}


@media print {
    .print {
        margin: 0;
        page-break-after: always;
        width: 100%;
    }
}
